#include <ruby.h>
#include <delaunay.h>
#include <narray.h>

struct TRIANGLEI {
    double px[3], py[3];
    int    id[3];
};
typedef struct TRIANGLEI trianglei;

extern VALUE cNArray;
VALUE cTriangleI;

static VALUE
trianglei_wrap_struct(VALUE klass, trianglei *tri)
{
    return( Data_Wrap_Struct(klass, 0, -1, tri) );
}

static VALUE
trianglei_new_s(double px[], double py[], int id[], VALUE klass)
{
    trianglei *tri;
    int i;
    tri = (trianglei *) malloc(sizeof(trianglei));
    for(i=0;i<3;i++){
	tri->px[i] = px[i];
	tri->py[i] = py[i];
	tri->id[i] = id[i];
    }
    return trianglei_wrap_struct(klass, tri);  
}

static VALUE
trianglei_new(VALUE klass, VALUE px, VALUE py, VALUE id)
{
    double cpx[3], cpy[3];
    int cid[3];
    int i;
    if ( (TYPE(px) == T_ARRAY) && (RARRAY(px)->len==3) ){
	for(i=0;i<3;i++){
	    cpx[i] = NUM2DBL( RARRAY(px)->ptr[i] );
	}
    } else {
	rb_raise(rb_eTypeError, "px must be a 3-element Array");
    }
    if (TYPE(py) == T_ARRAY && RARRAY(py)->len==3) {
	for(i=0;i<3;i++){
	    cpy[i] = NUM2DBL( RARRAY(py)->ptr[i] );
	}
    } else {
	rb_raise(rb_eTypeError, "py must be a 3-element Array");
    }
    if (TYPE(id) == T_ARRAY && RARRAY(id)->len==3) {
	for(i=0;i<3;i++){
	    cid[i] = NUM2INT( RARRAY(id)->ptr[i] );
	}
    } else {
	rb_raise(rb_eTypeError, "id must be a 3-element Array");
    }
    trianglei_new_s(cpx, cpy, cid, klass);
}

/*  cι¤ΤRubyMethood  */
static VALUE
trianglei_slice(VALUE self, VALUE i)
{
    struct TRIANGLEI *tri;
    int idx;
    double pxi, pyi;
    Data_Get_Struct(self, trianglei, tri);
    idx = NUM2INT(i);
    pxi = tri->px[idx];
    pyi = tri->py[idx];
    return( rb_ary_new3(2, rb_float_new(pxi), rb_float_new(pyi) ) );
}

static VALUE
trianglei_px(VALUE self)
{
    struct TRIANGLEI *tri;
    VALUE px;
    int shape=3, i;
    double *cpx;
    struct NARRAY *na;
    Data_Get_Struct(self, trianglei, tri);
    px = na_make_object(NA_DFLOAT, 1, &shape, cNArray);
    GetNArray(px, na);
    cpx = (double *)NA_PTR(na, 0);
    for(i=0;i<3;i++){
	cpx[i] = tri->px[i];
    }
    return(px);
}

static VALUE
trianglei_py(VALUE self)
{
    struct TRIANGLEI *tri;
    VALUE py;
    int shape=3, i;
    double *cpy;
    struct NARRAY *na;
    Data_Get_Struct(self, trianglei, tri);
    py = na_make_object(NA_DFLOAT, 1, &shape, cNArray);
    GetNArray(py, na);
    cpy = (double *)NA_PTR(na, 0);
    for(i=0;i<3;i++){
	cpy[i] = tri->py[i];
    }
    return(py);
}

static VALUE
trianglei_px_cyclic4(VALUE self)
{
    struct TRIANGLEI *tri;
    VALUE px;
    int shape=4, i;
    double *cpx;
    struct NARRAY *na;
    Data_Get_Struct(self, trianglei, tri);
    px = na_make_object(NA_DFLOAT, 1, &shape, cNArray);
    GetNArray(px, na);
    cpx = (double *)NA_PTR(na, 0);
    for(i=0;i<3;i++){
	cpx[i] = tri->px[i];
    }
    cpx[3] = tri->px[0];
    return(px);
}

static VALUE
trianglei_py_cyclic4(VALUE self)
{
    struct TRIANGLEI *tri;
    VALUE py;
    int shape=4, i;
    double *cpy;
    struct NARRAY *na;
    Data_Get_Struct(self, trianglei, tri);
    py = na_make_object(NA_DFLOAT, 1, &shape, cNArray);
    GetNArray(py, na);
    cpy = (double *)NA_PTR(na, 0);
    for(i=0;i<3;i++){
	cpy[i] = tri->py[i];
    }
    cpy[3] = tri->py[0];
    return(py);
}

static VALUE
trianglei_ids(VALUE self)
{
    struct TRIANGLEI *tri;
    VALUE ids;
    int shape=3, i;
    int *cids;
    struct NARRAY *na;
    Data_Get_Struct(self, trianglei, tri);
    ids = na_make_object(NA_LINT, 1, &shape, cNArray);
    GetNArray(ids, na);
    cids = (int *)NA_PTR(na, 0);
    for(i=0;i<3;i++){
	cids[i] = tri->id[i];
    }
    return(ids);
}

/* triangle  trianglei */
static VALUE
triangle2trianglei(triangle *t, double *px, double *py)
{
    VALUE ary;
    trianglei *tri;
    triangle *w;
    int i, id;
    w = t;
    ary = rb_ary_new();
    do {
	tri = (trianglei *) malloc(sizeof(trianglei));
	for(i=0;i<3;i++){
	    id = w->id[i];
	    tri->id[i] = id;
	    tri->px[i] = px[id];
	    tri->py[i] = py[id];
	}
	rb_ary_push( ary, trianglei_wrap_struct(cTriangleI, tri) );
	w = w->next;
    } while (w != NULL);
    return(ary);
}

/* trianglei_obj2cdoubleary β */
static double *
na2cdoubleary(src, length)
    VALUE src;
    int *length;
{
    VALUE chk;
    int len, i;
    double *rtn;
    double *ptr;
    struct NARRAY *na;

    chk = rb_obj_is_kind_of(src, cNArray);
    if (chk == Qfalse) {
        rb_raise(rb_eTypeError, "expect NArray");
    }

    src = na_cast_object(src, NA_DFLOAT);
    GetNArray(src, na);
    len = na->total;
    *length = len;
    ptr = (double *)NA_PTR(na, 0);

    rtn = ALLOC_N(double, len+3);

    for (i = 0; i < len; i++) {
	rtn[i] = (double)ptr[i];
    }
    
    return rtn;
}

/* trianglei_obj2cdoubleary β */
static double *
ary2cdoubleary(src, length)
    VALUE src;
    int *length;
{
    VALUE *ptr;
    int len, i;
    double *rtn;

    Check_Type(src, T_ARRAY);

    len = RARRAY(src)->len;
    *length = len;
    ptr = RARRAY(src)->ptr;

    rtn = ALLOC_N(double, len);

    for (i = 0; i < len; i++) {
	rtn[i] = NUM2DBL(ptr[i]);
    }

    return rtn;
}

/* NArray or Array  double  pointer  */
static double *
trianglei_obj2cdoubleary(src, length)
    VALUE src;
    int *length;
{
    VALUE chk;
    double *rtn;
    char *klass;
    VALUE rmiss;

    switch (TYPE(src)){
    case T_DATA:
        chk = rb_obj_is_kind_of(src, cNArray);
        rtn = na2cdoubleary(src, length);
        break;
    case T_ARRAY:
        rtn = ary2cdoubleary(src, length);
        break;
    default:
        rb_raise(rb_eTypeError, "expect double array");
        break;
    }

    return rtn;
}


static VALUE
trianglei_delaunay(VALUE klass, VALUE px, VALUE py)
{
    double *cpx, *cpy;
    int lenpx, lenpy, n, i;
    triangle *t;
    VALUE d_trianglei;

    cpx = trianglei_obj2cdoubleary(px, &lenpx);
    cpy = trianglei_obj2cdoubleary(py, &lenpy);
    if (lenpx != lenpy){
        rb_raise(rb_eArgError, "lengths of px and py must be the same");
    }
    n = lenpx;
    t = Delaunay2D(n, cpx, cpy);
    /* printf("###Delaunay2D done####\n");
       print_triangles("Delaunay2D",t); */
    d_trianglei = triangle2trianglei(t, cpx, cpy);
    free_triangle(t);
    return( d_trianglei );
}

void
Init_trianglei()
{
    cTriangleI = rb_define_class("TriangleI",rb_cObject);
    rb_define_singleton_method(cTriangleI,"new",trianglei_new,3);
    rb_define_method(cTriangleI, "[]", trianglei_slice, 1);
    rb_define_singleton_method(cTriangleI,"delaunay",trianglei_delaunay,2);
    rb_define_method(cTriangleI, "px", trianglei_px, 0);
    rb_define_method(cTriangleI, "py", trianglei_py, 0);
    rb_define_method(cTriangleI, "px_cyclic4", trianglei_px_cyclic4, 0);
    rb_define_method(cTriangleI, "py_cyclic4", trianglei_py_cyclic4, 0);
    rb_define_method(cTriangleI, "ids", trianglei_ids, 0);
}
